from multiprocessing import Process, Event


class Actor(Process):
    def __init__(self, **kwargs):
        Process.__init__(self)
        self.exit = Event()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        return self

    def terminate(self):
        self.exit.set()

    def stop(self):
        self.terminate()


class MyProcess(Actor):
    pass



from json import loads
import logging
from tornado import ioloop
from zmq.eventloop.zmqstream import ZMQStream

try:
    from snappy import decompress, UncompressError

except ImportError:
    print("Requires google snappy compression library")
    raise SystemExit

try:
    import zmq

except ImportError:
    print("Requires pyzmq")
    raise SystemExit

from cifsdk.zmq.socket import Context
from cifsdk.constants import REMOTE
from cifsdk.zmq.msg import Msg
from .constants import TRACE, RCVTIMEO, SNDTIMEO, LINGER

logger = logging.getLogger(__name__)

logger.setLevel(logging.ERROR)

if TRACE:
    logger.setLevel(logging.DEBUG)

basestring = (str, bytes)


class Base(object):

    def __init__(self, **kwargs):
        self.context = Context()
        self.socket = self.context.socket(zmq.REQ)
        self.nowait = kwargs.get('nowait', False)
        if self.nowait:
            self.socket = self.context.socket(zmq.DEALER)

        self.socket.RCVTIMEO = RCVTIMEO
        self.socket.SNDTIMEO = SNDTIMEO
        self.socket.LINGER = LINGER

        self.autoclose = kwargs.get('autoclose', False)

        self.remote = kwargs.get('remote', REMOTE)

        self.fireball = kwargs.get('fireball', False)

        self.loop = kwargs.get('loop')
        self.is_connected = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):

        # do our best to clean up potentially leaky FDs
        if hasattr(self, 'stream') and not self.stream.closed():
            self.stream.close()

        if hasattr(self, 'loop') and self.loop is not None:
            try:
                self.loop.close()
            except KeyError:
                pass

        self.socket.close()
        self.context.term()

    def _handle_message_fireball(self, m):
        logger.debug('message received')

        m = self._check_recv(loads(Msg.from_frame(m).data))

        self.response.append(m)

        self.num_responses -= 1
        logger.debug('num responses remaining: %i' % self.num_responses)

        if self.num_responses == 0:
            logger.debug('finishing up...')
            self.loop.stop()

    def _fireball_timeout(self):
        logger.info('fireball timeout')
        self.loop.stop()

    def _send_fireball(self, data, f_size):
        if len(data) == 0:
            logger.error('no data to send')
            return []

        self.loop = ioloop.IOLoop()
        self.socket.close()

        self.socket = self.context.socket(zmq.DEALER)
        self.socket.connect(self.remote)

        self.stream = ZMQStream(self.socket, io_loop=self.loop)
        self.stream.on_recv(self._handle_message_fireball)

        self.stream.io_loop.call_later(SNDTIMEO, self._fireball_timeout)

        self.response = []

        if not isinstance(data, list):
            data = [data]

        self.num_responses = int((len(data) / f_size))
        if (len(data) % f_size) != 0:
            self.num_responses += 1

        logger.debug('responses expected: %i' % self.num_responses)

        batch = []
        for d in data:
            batch.append(d)
            if len(batch) == f_size:
                self.socket.send_msg(
                    Msg(mtype=Msg.INDICATORS_CREATE, data=batch)
                )
                batch = []

        if len(batch):
            self.socket.send_msg(Msg(mtype=Msg.INDICATORS_CREATE, data=batch))

        logger.debug("starting loop to receive")
        self.loop.start()

        # clean up FDs
        self.loop.close()
        self.stream.close()
        self.socket.close()
        return self.response

    def _check_recv(self, data):
        if data.get('message') != 'success':
            if self.loop:
                self.loop.stop()

        if data.get('message') == 'unauthorized':
            raise PermissionError

        if data.get('message') == 'busy':
            raise ConnectionRefusedError('System Busy')

        if data.get('message') == 'invalid search':
            raise TypeError('invalid search')

        if data.get('status') != 'success':
            self.loop.stop()
            raise RuntimeError(data.get('message'))

        if data.get('data') is None:
            raise RuntimeError('invalid response')

        return data

    def _recv(self, decode=True, close=True):
        m = self.socket.recv_msg()

        if close:
            self.socket.close()

        if not decode:
            return m.data

        data = self._check_recv(loads(m.data))

        if isinstance(data.get('data'), bool):
            return data['data']

        try:
            data['data'] = decompress(data['data'])
        except (TypeError, UncompressError):
            pass

        return data.get('data')

    def _send(self, mtype, data='[]', nowait=False, decode=True):

        if not self.is_connected:
            self.socket.connect(self.remote)

        self.is_connected = True

        if isinstance(data, str):
            data = data.encode('utf-8')

        self.socket.send_msg(Msg(mtype=mtype, data=data))

        if self.nowait or nowait:
            if self.autoclose:
                self.socket.close()
            return

        return self._recv(decode=decode, close=self.autoclose)



import pytest


from cifsdk.client.zeromq import ZMQ as Client

@pytest.fixture
def client():
    yield Client()


def test_zmq(client):
    assert client.remote


def test_zmq_create(client):
    def _send(*args, **kwargs):
        return []

    client._send = _send

    rv = client.indicators_create(['asdfasdf'])
    assert rv == []


def test_zmq_search(client):
    def _send(*args, **kwargs):
        return []

    client._send = _send

    rv = client.indicators_search(['asdfasdf'])
    assert rv == []


def test_zmq_delete(client):
    def _send(*args, **kwargs):
        return []

    client._send = _send

    rv = client.indicators_delete(['asdfasdf'])
    assert rv == []



from setuptools import setup, find_packages
import os
import sys
import versioneer


# https://www.pydanny.com/python-dot-py-tricks.html
if sys.argv[-1] == 'test':
    test_requirements = [
        'pytest',
        'coverage',
        'pytest_cov',
    ]
    try:
        modules = map(__import__, test_requirements)

    except ImportError as e:
        err_msg = e.message.replace("No module named ", "")
        msg = "%s is not installed. Install your test requirments." % err_msg
        raise ImportError(msg)

    r = os.system('py.test test -v --cov=cifsdk --cov-fail-under=65 && '
                  'xenon -b B -m A -a A -e "cifsdk/_version*,cifsdk/client/http/utils.py,cifsdk/cli.py,cifsdk/client/zeromq/base.py" cifsdk')
    if r == 0:
        raise SystemExit

    raise RuntimeError('tests failed')

setup(
    name="cifsdk",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="CIFv5 SDK",
    long_description="CIFv5 Software Development Kit",
    url="https://github.com/csirtgadgets/cifsdk-v5",
    license='MPL2',
    classifiers=[
        "Topic :: System :: Networking",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "Programming Language :: Python"
    ],
    keywords=['network', 'security'],
    author="Wes Young",
    author_email="wes@csirtgadgets.com",
    packages=find_packages(exclude='test'),
    install_requires=[
        'ujson',
        'tornado',
        'msgpack',
        'pyzmq',
        'python-snappy',
        'arrow',
        'csirtg-dt',
        'csirtg-re',
        'csirtg-peers',
        'csirtg-spamhaus',
        'csirtg-indicator',
        'csirtgsdk',
    ],
    entry_points={
        'console_scripts': [
            'cif=cifsdk.cli:main'
        ]
    },
    test_suite="test"
)



from msgpack import unpackb, packb
import json
try:
    from snappy import compress, decompress

except ImportError:
    print("Requires google snappy compression lib to be installed..")
    raise

MAP = {
    1: 'indicators_create',
    2: 'indicators_search',
    3: 'indicators_delete'
}


def pack(data):
    if type(data) not in [str, bytes, dict, list]:
        raise TypeError(f'data must be of stdlib type, not: {type(data)}')

    return compress(packb(data))


def unpack(data, raw=False):
    return unpackb(decompress(data), raw=raw)


def mtype_to_int(mtype):
    for m in MAP:
        if MAP[m] == mtype:
            return m


class DecimalDecoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, float):
            return str(o)

        return super(DecimalDecoder, self).default(o)





from csirtg_indicator import Indicator
from cifsdk.client.http import HTTP as Client


def search(filters={}):
    if len(filters) == 0:
        filters = {
            'itype': 'ipv4',
            'tags': 'botnet'
        }

    return Client().search(filters)


def create(**kv):
    i = Indicator(**kv)

    return Client().indicators_create(i)



import os

TRACE = os.environ.get('CIFSDK_CLIENT_HTTP_TRACE')
TIMEOUT = os.getenv('CIFSDK_CLIENT_HTTP_TIMEOUT', 120)
RETRIES = os.getenv('CIFSDK_CLIENT_HTTP_RETRIES', 5)
RETRIES_DELAY = os.getenv('CIFSDK_CLIENT_HTTP_RETRIES_DELAY', '30,60')




from cifsdk.zmq.msg import Msg

import zmq


class Socket(zmq.Socket):

    def __init__(self, *args, **kwargs):
        zmq.Socket.__init__(self, *args, **kwargs)

        self.LINGER = 3000
        self.SNDTIMEO = 5000
        self.RCVTIMEO = 5000
        self.set_hwm(250)

    def send_msg(self, m, flags=0):
        if isinstance(m, Msg):
            m = m.to_frame()

        return self.send_multipart(m)

    def recv_msg(self, relay=False):
        m = self.recv_multipart()

        if relay:
            assert isinstance(relay, zmq.Socket) or isinstance(relay, Socket)
            return relay.send_msg(m)

        return Msg().from_frame(m)

    def reply(self, m, data=None):
        if data:
            m.data = data

        return self.send_msg(m)


class Context(zmq.Context):
    _socket_class = Socket


def main():
    ctx = Context()
    req = ctx.socket(zmq.REQ)
    rep = ctx.socket(zmq.REP)

    rep.bind('inproc://a')
    req.connect('inproc://a')
    A = [1, 2, 3, 4, 5, 6]

    # send/recv with pickle+zip
    req.send_compressed(A)
    B = rep.recv_compressed()

    print("Checking zipped pickle...")
    print(A == B)

    from pprint import pprint
    pprint(A)


if __name__ == '__main__':
    main()